﻿module InferenceModule

open System
open PuzzleModule
open System.Collections.Generic

type Inference() = class

    // Allows for mutable data
    let mutable MaxLoopCnt = 0
    let mutable LogicCnt = 0
    let mutable GuessCnt = 0
    let mutable BackTrackCnt = 0
    let mutable BackTrackingFlag = false
    let Arr = [| for a in 1 .. PuzzleModule.SIZE do yield 0 |]
    let PuzzleStack = new Stack<Puzzle>()
    
    member x.stack = PuzzleStack

    member x.maxLoopCnt 
        with get() = MaxLoopCnt
        and set(value) = MaxLoopCnt <- value

    member x.logicCnt 
        with get() = LogicCnt
        and set(value) = LogicCnt <- value

    member x.guessCnt 
        with get() = GuessCnt
        and set(value) = GuessCnt <- value

    member x.backTrackCnt 
        with get() = BackTrackCnt
        and set(value) = BackTrackCnt <- value

    member x.backTrackingFlag 
        with get() = BackTrackingFlag
        and set(value) = BackTrackingFlag <- value

    member x.arr = Arr

    // Function: Solve()
    member x.Solve(puzzle:Puzzle) = 
        let mutable retValue = true

        x.logicCnt <- 0
        x.guessCnt <- 0
        x.backTrackCnt <- 0
        x.backTrackingFlag <- false
        x.stack.Clear()
        x.maxLoopCnt <- 0

        let mutable thinkFlag = false
        let mutable continueLoop = true
        while (x.maxLoopCnt < 250000) && continueLoop do
            // Update the domains and sort in ascending order
            puzzle.UpdateRowColumnDomains()
            puzzle.UpdateSquareDomain()
            puzzle.UpdateGridDomain()
            x.Sort(puzzle)

            // Check the grid domain for inconsistency.  If there is a conflict, then 
            // backtrack, or the puzzle has no solution
            if x.IsInConsistent(puzzle) then
                x.backTrackingFlag <- true
                x.BackTracking(puzzle)

            // Verify that the puzzle is solved
            if x.IsSolved(puzzle) then
                retValue <- true
                continueLoop <- false
            else
                // Check the grid domains.  If a domain has only one number left then insert
                // it into the puzzle grid
                thinkFlag <- true
                let mutable i = 0
                while (i < GRID_SIZE) && thinkFlag do
                    if x.IsLastNumInDomain(puzzle.gDom.[i]) then
                        puzzle.InsertGrid(i, puzzle.gDom.[i].[8])
                        x.logicCnt <- x.logicCnt + 1
                        thinkFlag <- false
                    i <- i + 1

                // If the remaining grid domains have more than one number left, then we need 
                // to take a guess.  Taking a guess could result in no solution, which then
                // requires backtracking.  This code does not use lookahead mechanism
                if thinkFlag then
                    x.Think(puzzle)

            x.maxLoopCnt <- x.maxLoopCnt + 1
        retValue

    // Function: IsInConsistent()
    member x.IsInConsistent(puzzle:Puzzle) =
        let mutable retval = false
        let mutable continueMethod = true
        let mutable i = 0
        let mutable j = 0
        let mutable r = 0
        let mutable c = 0

        // Verify grid vs gDom configuration 
        for r = 0 to SIZE-1 do 
            for c = 0 to SIZE-1 do 
                if puzzle.gDom.[(r * 9) + c].[8] = 0 && puzzle.grid.[r].[c] = 0 then
                    retval <- true
                  
        if retval = false then 
            // Verify row configuration
            let mutable row = 0
            while (row < SIZE) && continueMethod do
               for i = 0 to SIZE-1 do
                  if x.IsLastNumInDomain(puzzle.gDom.[(row*SIZE) +  i]) then
                     x.arr.[i] <- puzzle.gDom.[(row*SIZE) + i].[8]
                  else
                     x.arr.[i] <- 0

               i <- 0
               while (i < SIZE) && continueMethod do
                  if x.arr.[i] = 0 then
                     continueMethod <- true
                  else
                     j <- 0
                     while (j < SIZE) && continueMethod do
                        if not(i=j) && (x.arr.[i] = x.arr.[j]) then
                           retval <- true
                           continueMethod <- false
                        j <- j+1
                  i <- i+1
               row <- row+1

            // Verify column configuration
            let mutable col = 0
            while (col < SIZE) && continueMethod do
               if x.IsLastNumInDomain(puzzle.gDom.[0+col]) then
                  x.arr.[0] <- puzzle.gDom.[0+col].[8]
               else
                  x.arr.[0] <- 0
                    
               if x.IsLastNumInDomain(puzzle.gDom.[9+col]) then
                  x.arr.[1] <- puzzle.gDom.[9+col].[8]
               else
                  x.arr.[1] <- 0

               if x.IsLastNumInDomain(puzzle.gDom.[18+col]) then
                  x.arr.[2] <- puzzle.gDom.[18+col].[8]
               else
                  x.arr.[2] <- 0
                        
               if x.IsLastNumInDomain(puzzle.gDom.[27+col]) then
                  x.arr.[3] <- puzzle.gDom.[27+col].[8]
               else
                  x.arr.[3] <- 0
                        
               if x.IsLastNumInDomain(puzzle.gDom.[36+col]) then
                  x.arr.[4] <- puzzle.gDom.[36+col].[8]
               else
                  x.arr.[4] <- 0
                        
               if x.IsLastNumInDomain(puzzle.gDom.[45+col]) then
                  x.arr.[5] <- puzzle.gDom.[45+col].[8]
               else
                  x.arr.[5] <- 0
                        
               if x.IsLastNumInDomain(puzzle.gDom.[54+col]) then
                  x.arr.[6] <- puzzle.gDom.[54+col].[8]
               else
                  x.arr.[6] <- 0
                        
               if x.IsLastNumInDomain(puzzle.gDom.[63+col]) then
                  x.arr.[7] <- puzzle.gDom.[63+col].[8]
               else
                  x.arr.[7] <- 0
                        
               if x.IsLastNumInDomain(puzzle.gDom.[72+col]) then
                  x.arr.[8] <- puzzle.gDom.[72+col].[8]
               else
                  x.arr.[8] <- 0

               i <- 0
               while (i < SIZE) && continueMethod do
                  if not(x.arr.[i] = 0) then
                     j <- 0
                     while (j < SIZE) && continueMethod do
                        if not(i = j) && (x.arr.[i] = x.arr.[j]) then
                           retval <- true
                           continueMethod <- false
                        j <- j+1
                  i <- i+1
               col <- col+1

        retval

    // Function: BackTracking()
    member x.BackTracking(puzzle:Puzzle) =
        x.Restore(puzzle)
        x.stack.Pop() |> ignore
        x.backTrackCnt <- x.backTrackCnt+1

        if puzzle.domainIndex < 8 then
            // Remove the last number that was guessed from the last domain
            puzzle.rmD(puzzle.gDom.[puzzle.gridIndex], puzzle.gridValue)

            // Use the next number in the domain
            puzzle.domainIndex <- puzzle.domainIndex+1
            puzzle.gridValue <- puzzle.gDom.[puzzle.gridIndex].[puzzle.domainIndex]
            x.ArchivePuzzle(puzzle)
            puzzle.InsertGrid(puzzle.gridIndex, puzzle.gridValue)
            x.guessCnt <- x.guessCnt+1
        ()

    // Function: Restore
    member x.Restore(puzzle:Puzzle) =
        let p = x.stack.Peek()

        for i = 0 to SIZE-1 do
            for j = 0 to SIZE-1 do
                puzzle.grid.[i].[j] <- p.grid.[i].[j]
                puzzle.row.[i].[j] <- p.row.[i].[j]
                puzzle.col.[i].[j] <- p.col.[i].[j]
                puzzle.sqr.[i].[j] <- p.sqr.[i].[j]

        for i = 0 to GRID_SIZE-1 do
            for j = 0 to SIZE-1 do
                puzzle.gDom.[i].[j] <- p.gDom.[i].[j]

        puzzle.gridIndex <- p.gridIndex
        puzzle.gridValue <- p.gridValue
        puzzle.domainIndex <- p.domainIndex
        ()

    // Function: Think()
    member x.Think(puzzle:Puzzle) =
        // Archive the puzzle state to provide for backtracking
        puzzle.gridIndex <- x.GetNextGridIndex(puzzle)
        puzzle.gridValue <- x.GetNextGridDomainValue(puzzle)
        x.ArchivePuzzle(puzzle)

        // Insert the number into the puzzle grid
        puzzle.InsertGrid(puzzle.gridIndex, puzzle.gridValue)
        x.guessCnt <- x.guessCnt+1

        ()

    // Function: IsLastNumInDomain()
    member x.IsLastNumInDomain(array:int[]) =
        let mutable retval = false
        if (array.[7] = 0) && (array.[8] > 0) then
            retval <- true
        retval

    // Function: Sort()
    member x.Sort(puzzle:Puzzle) =
        for i = 0 to SIZE-1  do
            System.Array.Sort(puzzle.row.[i])
            System.Array.Sort(puzzle.col.[i])
            System.Array.Sort(puzzle.sqr.[i])

        for i = 0 to GRID_SIZE-1 do
            System.Array.Sort(puzzle.gDom.[i])
        ()

    // Function: IsSolved()
    member x.IsSolved(puzzle:Puzzle) =
        let mutable retval = true
        let mutable i = 0
        let mutable j = 0
        let mutable continueMethod = true
        while (i < SIZE) && continueMethod do
            j <- 0
            while (j < SIZE) && continueMethod do
                if not(puzzle.row.[i].[j] = 0) then
                    retval <- false
                    continueMethod <- false
                if not(puzzle.col.[i].[j] = 0) then
                    retval <- false
                    continueMethod <- false
                j <- j+1
            i <- i+1
        retval

    // Function: ArchivePuzzle()
    member x.ArchivePuzzle(puzzle:Puzzle) =
        let p = new Puzzle()
        for i = 0 to SIZE-1 do
            for j = 0 to SIZE-1 do
                p.grid.[i].[j] <- puzzle.grid.[i].[j]
                p.row.[i].[j] <- puzzle.row.[i].[j]
                p.col.[i].[j] <- puzzle.col.[i].[j]
                p.sqr.[i].[j] <- puzzle.sqr.[i].[j]

        for i = 0 to GRID_SIZE-1 do
            for j = 0 to SIZE-1 do
                p.gDom.[i].[j] <- puzzle.gDom.[i].[j]

        p.gridIndex <- puzzle.gridIndex
        p.domainIndex <- puzzle.domainIndex
        p.gridValue <- puzzle.gridValue

        x.stack.Push(p)
        ()

    // Function: GetNextGridIndex()
    member x.GetNextGridIndex(puzzle:Puzzle) =
        let mutable retval = -1
        let mutable i = 0
        let mutable continueMethod = true
        while (i < GRID_SIZE) && continueMethod do
            if not(puzzle.gDom.[i].[8] = 0) then
                retval <- i
                continueMethod <- false
            i <- i+1
        retval

    // Function: GetNextGridDomainValue()
    member x.GetNextGridDomainValue(puzzle:Puzzle) =
        let mutable retval = -1
        let mutable i = 0
        let mutable continueMethod = true
        while (i < SIZE) && continueMethod do
            if not(puzzle.gDom.[puzzle.gridIndex].[i] = 0) then
                puzzle.domainIndex <- i
                retval <- puzzle.gDom.[puzzle.gridIndex].[i]
                continueMethod <- false
            i <- i+1
        retval

    member x.Statistics(puzzle:Puzzle, buf:string) =
        // TODO
        ()
end