﻿module PuzzleModule

open System

let SIZE = 9
let GRID_SIZE = 81

type Puzzle()  = class

    // Allows for mutable data
    let mutable GridIndex = 0
    let mutable DomainIndex = 0
    let mutable GridValue = 0
    let Grid = [| for a in 1 .. SIZE do yield [| 1 .. SIZE |] |]
    let GDom = [| for a in 1 .. GRID_SIZE do yield [| 1 .. SIZE |] |]
    let Row =  [| for a in 1 .. SIZE do yield [| 1 .. SIZE |] |]
    let Col = [| for a in 1 .. SIZE do yield [| 1 .. SIZE |] |]
    let Sqr = [| for a in 1 .. SIZE do yield [| 1 .. SIZE |] |]

    member x.gridIndex 
        with get() = GridIndex
        and set(value) = GridIndex <- value

    member x.domainIndex 
        with get() = DomainIndex
        and set(value) = DomainIndex <- value

    member x.gridValue 
        with get() = GridValue
        and set(value) = GridValue <- value

    member x.grid = Grid

    member x.gDom = GDom

    member x.row = Row
    member x.col = Col
    member x.sqr = Sqr

    // Function: Initialize
    member x.Initialize() = 
        for i = 0 to SIZE-1 do
            for j = 0 to SIZE-1 do
                x.grid.[i].[j] <- 0
                x.row.[i].[j] <- j + 1
                x.col.[i].[j] <- j + 1

        for i = 0 to SIZE-1 do
            x.sqr.[0].[i] <- i+1
            x.sqr.[1].[i] <- i+1
            x.sqr.[2].[i] <- i+1
            x.sqr.[3].[i] <- i+1
            x.sqr.[4].[i] <- i+1
            x.sqr.[5].[i] <- i+1
            x.sqr.[6].[i] <- i+1
            x.sqr.[7].[i] <- i+1
            x.sqr.[8].[i] <- i+1

        for i = 0 to GRID_SIZE-1 do
            x.gDom.[i].[0] <- 1
            x.gDom.[i].[1] <- 2
            x.gDom.[i].[2] <- 3
            x.gDom.[i].[3] <- 4
            x.gDom.[i].[4] <- 5
            x.gDom.[i].[5] <- 6
            x.gDom.[i].[6] <- 7
            x.gDom.[i].[7] <- 8
            x.gDom.[i].[8] <- 9
        ()

    // Function: IsInDomain
    member x.IsInDomain(array:int[], value:int) = 
        let mutable inDomain = false
        let mutable i = 0
        let mutable continueMethod = true
        while (i < array.Length) && continueMethod do
            if array.[i] = value then
                inDomain <- true
                continueMethod <- false
            i <- i+1
        inDomain
    
    // Function: rmD
    member x.rmD(array:int[], value:int) = 
        for i = 0 to array.Length-1 do
            if array.[i] = value then
                array.[i] <- 0
        ()

    // Function: InsertGrid
    member x.InsertGrid(k, value) = 
        let r = k / 9;
        let index = k - (r * 9)
        x.grid.[r].[index] <- value
        ()
    
    // Function: UpdateRowColumnDomains
    member x.UpdateRowColumnDomains() = 
        // Check the rows & columns
        for i = 0 to SIZE-1 do
            for j = 0 to SIZE-1 do
                if not(x.grid.[i].[j] = 0) then
                    x.rmD(x.row.[i], x.grid.[i].[j])
                    x.rmD(x.col.[j], x.grid.[i].[j])
        ()

    member x.UpdateSquareDomain() = 
        // Check Square #0
        for i = 0 to 2 do
            for j = 0 to 2 do
                if not(x.grid.[i].[j] = 0) then
                    x.rmD(x.sqr.[0], x.grid.[i].[j])
                    
        // Check Square #1
        for i = 0 to 2 do
            for j = 3 to 5 do
                if not(x.grid.[i].[j] = 0) then
                    x.rmD(x.sqr.[1], x.grid.[i].[j])
                    
        // Check Square #2
        for i = 0 to 2 do
            for j = 6 to 8 do
                if not(x.grid.[i].[j] = 0) then
                    x.rmD(x.sqr.[2], x.grid.[i].[j])
                    
        // Check Square #3
        for i = 3 to 5 do
            for j = 0 to 2 do
                if not(x.grid.[i].[j] = 0) then
                    x.rmD(x.sqr.[3], x.grid.[i].[j])
                    
        // Check Square #4
        for i = 3 to 5 do
            for j = 3 to 5 do
                if not(x.grid.[i].[j] = 0) then
                    x.rmD(x.sqr.[4], x.grid.[i].[j])
                    
        // Check Square #5
        for i = 3 to 5 do
            for j = 6 to 8 do
                if not(x.grid.[i].[j] = 0) then
                    x.rmD(x.sqr.[5], x.grid.[i].[j])
                    
        // Check Square #6
        for i = 6 to 8 do
            for j = 0 to 2 do
                if not(x.grid.[i].[j] = 0) then
                    x.rmD(x.sqr.[6], x.grid.[i].[j])
                    
        // Check Square #7
        for i = 6 to 8 do
            for j = 3 to 5 do
                if not(x.grid.[i].[j] = 0) then
                    x.rmD(x.sqr.[7], x.grid.[i].[j])
                    
        // Check Square #8
        for i = 6 to 8 do
            for j = 6 to 8 do
                if not(x.grid.[i].[j] = 0) then
                    x.rmD(x.sqr.[8], x.grid.[i].[j])
        ()

    member x.UpdateGridDomain() = 
        // If the grid has an entry the grid domain is set to zero
        for i = 0 to SIZE-1 do
            for j = 0 to SIZE-1 do
                if not(x.grid.[i].[j] = 0) then
                    for k = 0 to SIZE - 1 do
                        x.gDom.[(i*SIZE) + j].[k] <- 0

        // If the grid number is not in the row then remove it from the grid domain
        for i = 0 to SIZE-1 do
            for j = 1 to SIZE do
                if not(x.IsInDomain(x.row.[i], j)) then
                    x.rmD(x.gDom.[(i*9)+0], j)
                    x.rmD(x.gDom.[(i*9)+1], j)
                    x.rmD(x.gDom.[(i*9)+2], j)
                    x.rmD(x.gDom.[(i*9)+3], j)
                    x.rmD(x.gDom.[(i*9)+4], j)
                    x.rmD(x.gDom.[(i*9)+5], j)
                    x.rmD(x.gDom.[(i*9)+6], j)
                    x.rmD(x.gDom.[(i*9)+7], j)
                    x.rmD(x.gDom.[(i*9)+8], j)

        // If the grid number is not in the column then remove it from the grid domain
        for i = 0 to SIZE-1 do
            for j = 1 to SIZE do
                if not(x.IsInDomain(x.col.[i], j)) then
                    x.rmD(x.gDom.[i], j)
                    x.rmD(x.gDom.[(1*9)+i], j)
                    x.rmD(x.gDom.[(2*9)+i], j)
                    x.rmD(x.gDom.[(3*9)+i], j)
                    x.rmD(x.gDom.[(4*9)+i], j)
                    x.rmD(x.gDom.[(5*9)+i], j)
                    x.rmD(x.gDom.[(6*9)+i], j)
                    x.rmD(x.gDom.[(7*9)+i], j)
                    x.rmD(x.gDom.[(8*9)+i], j)

        // If the grid number is not in the square domain then remove it from the grid domain
        for i = 1 to SIZE do
            if not(x.IsInDomain(x.sqr.[0], i)) then
                    x.rmD(x.gDom.[0], i)
                    x.rmD(x.gDom.[1], i)
                    x.rmD(x.gDom.[2], i)
                    x.rmD(x.gDom.[9], i)
                    x.rmD(x.gDom.[10], i)
                    x.rmD(x.gDom.[11], i)
                    x.rmD(x.gDom.[18], i)
                    x.rmD(x.gDom.[19], i)
                    x.rmD(x.gDom.[20], i)
                    
        for i = 1 to SIZE do
            if not(x.IsInDomain(x.sqr.[1], i)) then
                    x.rmD(x.gDom.[3], i)
                    x.rmD(x.gDom.[4], i)
                    x.rmD(x.gDom.[5], i)
                    x.rmD(x.gDom.[12], i)
                    x.rmD(x.gDom.[13], i)
                    x.rmD(x.gDom.[14], i)
                    x.rmD(x.gDom.[21], i)
                    x.rmD(x.gDom.[22], i)
                    x.rmD(x.gDom.[23], i)
                    
        for i = 1 to SIZE do
            if not(x.IsInDomain(x.sqr.[2], i)) then
                    x.rmD(x.gDom.[6], i)
                    x.rmD(x.gDom.[7], i)
                    x.rmD(x.gDom.[8], i)
                    x.rmD(x.gDom.[15], i)
                    x.rmD(x.gDom.[16], i)
                    x.rmD(x.gDom.[17], i)
                    x.rmD(x.gDom.[24], i)
                    x.rmD(x.gDom.[25], i)
                    x.rmD(x.gDom.[26], i)
                    
        for i = 1 to SIZE do
            if not(x.IsInDomain(x.sqr.[3], i)) then
                    x.rmD(x.gDom.[27], i)
                    x.rmD(x.gDom.[28], i)
                    x.rmD(x.gDom.[29], i)
                    x.rmD(x.gDom.[36], i)
                    x.rmD(x.gDom.[37], i)
                    x.rmD(x.gDom.[38], i)
                    x.rmD(x.gDom.[45], i)
                    x.rmD(x.gDom.[46], i)
                    x.rmD(x.gDom.[47], i)
                    
        for i = 1 to SIZE do
            if not(x.IsInDomain(x.sqr.[4], i)) then
                    x.rmD(x.gDom.[30], i)
                    x.rmD(x.gDom.[31], i)
                    x.rmD(x.gDom.[32], i)
                    x.rmD(x.gDom.[39], i)
                    x.rmD(x.gDom.[40], i)
                    x.rmD(x.gDom.[41], i)
                    x.rmD(x.gDom.[48], i)
                    x.rmD(x.gDom.[49], i)
                    x.rmD(x.gDom.[50], i)
                    
        for i = 1 to SIZE do
            if not(x.IsInDomain(x.sqr.[5], i)) then
                    x.rmD(x.gDom.[33], i)
                    x.rmD(x.gDom.[34], i)
                    x.rmD(x.gDom.[35], i)
                    x.rmD(x.gDom.[42], i)
                    x.rmD(x.gDom.[43], i)
                    x.rmD(x.gDom.[44], i)
                    x.rmD(x.gDom.[51], i)
                    x.rmD(x.gDom.[52], i)
                    x.rmD(x.gDom.[53], i)
                    
        for i = 1 to SIZE do
            if not(x.IsInDomain(x.sqr.[6], i)) then
                    x.rmD(x.gDom.[54], i)
                    x.rmD(x.gDom.[55], i)
                    x.rmD(x.gDom.[56], i)
                    x.rmD(x.gDom.[63], i)
                    x.rmD(x.gDom.[64], i)
                    x.rmD(x.gDom.[65], i)
                    x.rmD(x.gDom.[72], i)
                    x.rmD(x.gDom.[73], i)
                    x.rmD(x.gDom.[74], i)
                    
        for i = 1 to SIZE do
            if not(x.IsInDomain(x.sqr.[7], i)) then
                    x.rmD(x.gDom.[57], i)
                    x.rmD(x.gDom.[58], i)
                    x.rmD(x.gDom.[59], i)
                    x.rmD(x.gDom.[66], i)
                    x.rmD(x.gDom.[67], i)
                    x.rmD(x.gDom.[68], i)
                    x.rmD(x.gDom.[75], i)
                    x.rmD(x.gDom.[76], i)
                    x.rmD(x.gDom.[77], i)
                    
        for i = 1 to SIZE do
            if not(x.IsInDomain(x.sqr.[8], i)) then
                    x.rmD(x.gDom.[60], i)
                    x.rmD(x.gDom.[61], i)
                    x.rmD(x.gDom.[62], i)
                    x.rmD(x.gDom.[69], i)
                    x.rmD(x.gDom.[70], i)
                    x.rmD(x.gDom.[71], i)
                    x.rmD(x.gDom.[78], i)
                    x.rmD(x.gDom.[79], i)
                    x.rmD(x.gDom.[80], i)

        ()

end